import configparser
import os
from typing import List

import skg_main.skg_mgrs.connector_mgr as conn
from skg_main.skg_mgrs.skg_reader import Skg_Reader
from skg_main.skg_model.schema import Timestamp as skg_Timestamp
from skg_main.skg_model.semantics import EntityForest

from sha_learning.case_studies.auto_twin.sul_functions import label_event, parse_data, get_rand_param, \
    is_chg_pt
from sha_learning.domain.lshafeatures import Event, ProbDistribution
from sha_learning.domain.sigfeatures import Timestamp as lsha_Timestamp
from sha_learning.domain.sulfeatures import SystemUnderLearning, RealValuedVar, FlowCondition
from sha_learning.learning_setup.teacher import Teacher

config = configparser.ConfigParser()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

CS = config['SUL CONFIGURATION']['CASE_STUDY']

DRIVER_SIG = ['s_id']
DEFAULT_M = 0
DEFAULT_DISTR = 0

args = {'name': 'auto_twin', 'driver': DRIVER_SIG, 'default_m': DEFAULT_M, 'default_d': DEFAULT_DISTR}

pov = config['AUTO-TWIN CONFIGURATION']['POV'].lower()


def foo_model(interval: List[lsha_Timestamp]):
    return interval


# define flow conditions
foo_fc: FlowCondition = FlowCondition(0, foo_model)

# define distributions
foo_distr = ProbDistribution(0, {'avg': 0.0})

model2distr = {0: [], 1: []}
s_id = RealValuedVar([foo_fc], [], model2distr, label='s_id')
if pov == 'plant':
    state_sig = RealValuedVar([foo_fc], [], model2distr, label='state_vec')

act_to_sensors = dict()
auto_twin_cs = SystemUnderLearning([], [], parse_data, label_event, get_rand_param, is_chg_pt, args=args)


def getSUL():
    # define events
    driver = conn.get_driver()
    reader: Skg_Reader = Skg_Reader(driver)
    unique_events = reader.get_activities()

    # FIXME should be generic
    if 'Dirty' in [e.act.split(' ')[0] for e in unique_events]:
        act_to_sensors = {"Dirty Material Input": "S1",
                          "Assembled production": "S2",
                          "Composition of charges": "S3",
                          "Sterilizer Load Released": "S4",
                          "Commissioner": "S5",
                          "Load L+D released": "S6",
                          "Assembly": "S7",
                          "Loaded on L+D trolley": "S8",
                          "L+D Load Started": "S9",
                          "Enter Cleaning": "S10",
                          "Enter Storage": "S11",
                          "Exit Storage": "S12"}
    elif 'Read Lock Status' in [e.act for e in unique_events]:
        act_to_sensors = {'Pass Sensor S1': 'S1', 'Pass Sensor S2': 'S2', 'Pass Sensor S3': 'S3',
                          'Pass Sensor S4': 'S4',
                          'Pass Sensor S5': 'S5', 'Pass Sensor S6': 'S6', 'Pass Sensor S101': 'S101',
                          'Pass Sensor S105': 'S105', 'Pass Sensor S100': 'S100', 'Pass Sensor S7': 'S7',
                          'Pass Sensor S8': 'S8', 'Pass Sensor S102': 'S102', 'Pass Sensor S104': 'S104',
                          'Pass Sensor S9': 'S9', 'Pass Sensor S10': 'S10', 'Pass Sensor S103': 'S103',
                          'Pass Sensor S11': 'S11', 'Pass Sensor S12': 'S12', 'Pass Sensor S13': 'S13',
                          'Pass Sensor S14': 'S14', 'Pass Sensor S15': 'S15', 'Pass Sensor S16': 'S16',
                          'Pass Sensor S17': 'S17', 'Start Break': 'S200', 'Stop Break': 'S201',
                          'Read Lock Status': 'S202', 'Read WIP amount': 'S203', 'Pass Sensor S106': 'S106',
                          'Pass Sensor CS001': 'S1', 'Pass Sensor CS002': 'S2', 'Pass Sensor CS003': 'S3',
                          'Pass Sensor CS004': 'S4', 'Pass Sensor CS005': 'S5', 'Pass Sensor CS006': 'S6',
                          'Pass Sensor CS101': 'S101', 'Pass Sensor CS105': 'S105', 'Pass Sensor CS100': 'S100',
                          'Pass Sensor CS007': 'S7', 'Pass Sensor CS008': 'S8', 'Pass Sensor CS102': 'S102',
                          'Pass Sensor CS104': 'S104', 'Pass Sensor CS009': 'S9', 'Pass Sensor CS010': 'S10',
                          'Pass Sensor CS4102': 'S4102', 'Pass Sensor CS4103': 'S4103',
                          'Pass Sensor CS4104': 'S4104', 'Pass Sensor CS4201': 'S4201',
                          'Pass Sensor CS4202': 'S4202', 'Pass Sensor CS4301': 'S4301',
                          'Pass Sensor CS4401': 'S4401', 'Pass Sensor CS106': 'S106', 'Pass Sensor CS011': 'S11',
                          'Pass Sensor CS012': 'S12', 'Pass Sensor CS013': 'S13', 'Pass Sensor CS014': 'S14',
                          'Pass Sensor CS015': 'S15', 'Pass Sensor CS016': 'S16', 'Pass Sensor CS4101': 'S4101',
                          'Pass Sensor CS4203': 'S4203', 'Pass Sensor CS4204': 'S4204', 'Pass Sensor VS001': 'S501',
                          'Pass Sensor VS002': 'S502', 'Pass Sensor VS003': 'S503',
                          'Read power': 'S601', 'Read temperature': 'S602'}
    else:
        act_to_sensors = {'Pass Sensor LOAD_1': 'S11', 'Pass Sensor PROCESS_1': 'S12', 'Pass Sensor UNLOAD_1': 'S13',
                          'Pass Sensor LOAD_2': 'S21', 'Pass Sensor PROCESS_2': 'S22', 'Pass Sensor UNLOAD_2': 'S23',
                          'Pass Sensor FAIL_1': 'S14', 'Pass Sensor BLOCK_2': 'S24', 'Pass Sensor LOAD_3': 'S31',
                          'Pass Sensor PROCESS_3': 'S32', 'Pass Sensor LOAD_4': 'S41', 'Pass Sensor UNLOAD_3': 'S33',
                          'Pass Sensor PROCESS_4': 'S42', 'Pass Sensor LOAD_5': 'S51', 'Pass Sensor PROCESS_5': 'S52',
                          'Pass Sensor UNLOAD_5': 'S53', 'Pass Sensor UNLOAD_4': 'S43', 'Pass Sensor FAIL_5': 'S54',
                          'Pass Sensor BLOCK_1': 'S15', 'Pass Sensor BLOCK_5': 'S55', 'Pass Sensor BLOCK_3': 'S34',
                          'Pass Sensor BLOCK_4': 'S44'}

    for e in unique_events:
        if e.act in act_to_sensors:
            e.act = act_to_sensors[e.act]

    events: List[Event] = [Event('', e.act, e.act.lower()) for e in
                           unique_events]

    if pov == 'plant':
        labels_hierarchy = reader.get_entity_labels_hierarchy()
        resources = reader.get_resources(labels_hierarchy=labels_hierarchy)
    else:
        resources = []

    vars = [s_id] if pov != 'plant' else [state_sig, s_id]
    auto_twin_cs = SystemUnderLearning(vars, events, parse_data, label_event, get_rand_param, is_chg_pt, args=args)

    conn.close_connection(driver)
    return auto_twin_cs, act_to_sensors


test = False
if test:
    driver = conn.get_driver()
    reader: Skg_Reader = Skg_Reader(driver)

    evt_seqs = []

    if 'START_T' in config['AUTO-TWIN CONFIGURATION'] and 'END_T' in config['AUTO-TWIN CONFIGURATION']:
        START_T = int(config['AUTO-TWIN CONFIGURATION']['START_T'])
        END_T = int(config['AUTO-TWIN CONFIGURATION']['END_T'])
    else:
        def parse_date(s: str):
            fields = s.split('-')
            return skg_Timestamp(int(fields[0]), int(fields[1]), int(fields[2]), int(fields[3]), int(fields[4]),
                                 int(fields[5]))


        START_T = parse_date(config['AUTO-TWIN CONFIGURATION']['START_DATE'])
        END_T = parse_date(config['AUTO-TWIN CONFIGURATION']['END_DATE'])

    if pov != 'plant':
        TEST_N = 17

        if config['AUTO-TWIN CONFIGURATION']['POV'].lower() == 'item':
            entities = reader.get_items(labels_hierarchy=reader.get_entity_labels_hierarchy(), limit=TEST_N,
                                        random=True)
        else:
            entities = reader.get_resources(labels_hierarchy=reader.get_resource_labels_hierarchy(), limit=TEST_N,
                                            random=True)

        for entity in entities[:TEST_N]:
            if pov == 'item':
                entity_tree = reader.get_entity_tree(entity.entity_id, EntityForest([]), reverse=True)
                events = reader.get_events_by_entity_tree_and_timestamp(entity_tree[0], START_T, END_T, pov)
            elif pov == 'resource':
                entity_tree = reader.get_entity_tree(entity.entity_id, EntityForest([]))
                events = reader.get_events_by_entity_tree_and_timestamp(entity_tree[0], START_T, END_T, pov)
            if len(events) > 0:
                evt_seqs.append(events)
    else:
        events = reader.get_events_by_timestamp(START_T, END_T)
        entity_tree = reader.get_entity_tree("Oven", EntityForest([]))
        events = reader.get_events_by_entity_tree_and_timestamp(entity_tree[0], START_T, END_T, pov)
        if len(events) > 0:
            evt_seqs.append(events)

    teacher = Teacher(auto_twin_cs)

    for seq in evt_seqs:
        auto_twin_cs.process_data(seq)
        print(auto_twin_cs.traces[-1])
        # auto_twin_cs.plot_trace(-1)
        id_cluster = teacher.ht_query(auto_twin_cs.traces[-1], foo_fc)
        print(id_cluster)

    conn.close_connection(driver)



import configparser
import os
from typing import List

import matplotlib.pyplot as plt
import numpy as np

from sha_learning.domain.lshafeatures import TimedTrace, Event
from sha_learning.domain.sigfeatures import SampledSignal

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

CS_VERSION = int(config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', '')[0])
SAVE_PATH = config['SUL CONFIGURATION']['SHA_SAVE_PATH']


def double_plot(signal1: SampledSignal, signal2: SampledSignal, signal3: SampledSignal, t: TimedTrace,
                title: str, events_cs: List[Event]):
    fig, axs = plt.subplots(3, figsize=(40, 30))

    timestamps1 = [pt.timestamp for pt in signal1.points]
    timestamps2 = [pt.timestamp for pt in signal2.points]
    timestamps3 = [pt.timestamp for pt in signal3.points]
    v1 = [pt.value for pt in signal1.points]
    v2 = [pt.value for pt in signal2.points]
    v3 = [pt.value for pt in signal3.points]

    t1 = [x.to_secs() for x in timestamps1]
    axs[0].plot(t1, v1, label='fatigue')
    axs[0].plot(t1, [0] * len(v1), '--', color='grey', linewidth=.5)

    t2 = [x.to_secs() for x in timestamps2]
    axs[1].plot(t2, v2, label='busy')
    axs[1].plot(t2, [0] * len(v2), '--', color='grey', linewidth=.5)

    t3 = [x.to_secs() for x in timestamps3]
    axs[2].plot(t3, v3, label='busy')
    axs[2].plot(t3, [0] * len(v3), '--', color='grey', linewidth=.5)

    HEIGHT1 = max(v1)
    HEIGHT2 = max(v2)
    HEIGHT3 = max(v3)

    colors = ['g', 'r', 'orange', 'purple']

    marker = 'x'
    height1 = HEIGHT1
    height2 = HEIGHT2
    height3 = HEIGHT3

    i = 0
    labels = [e.symbol for e in t.e]
    events = [ts.to_secs() for ts in t.t]
    axs[0].plot(events, [height1] * len(events), marker, color=colors[i], label=labels[i])
    axs[0].vlines(events, [0] * len(events), [height1] * len(events), color=colors[i], linewidth=0.5)
    for i, e in enumerate(events):
        text = [x.guard + ', ' + x.chan for x in events_cs if x.symbol == labels[i]][0]
        axs[0].text(e, height1, text, fontsize=20)

    i = 0
    axs[1].plot(events, [height2] * len(events), marker, color=colors[i], label=labels[i])
    axs[1].vlines(events, [0] * len(events), [height2] * len(events), color=colors[i], linewidth=0.5)
    for i, e in enumerate(events):
        text = [x.guard + ', ' + x.chan for x in events_cs if x.symbol == labels[i]][0]
        axs[1].text(e, height2, text, fontsize=20)

    i = 0
    axs[2].plot(events, [height3] * len(events), marker, color=colors[i], label=labels[i])
    axs[2].vlines(events, [0] * len(events), [height3] * len(events), color=colors[i], linewidth=0.5)
    for i, e in enumerate(events):
        text = [x.guard + ', ' + x.chan for x in events_cs if x.symbol == labels[i]][0]
        axs[2].text(e, height3, text, fontsize=20)

    step = 30
    xticks = [str(x.to_secs()) for x in timestamps1][::step]
    axs[0].set_xticks(ticks=t1[::step])
    axs[0].set_xticklabels(labels=xticks, fontsize=24)
    yticks = np.arange(0, max(v1), 0.05)
    axs[0].set_yticks(ticks=yticks)
    axs[0].set_yticklabels(labels=['{:.1f}'.format(l) for l in yticks], fontsize=24)

    step = 30
    xticks = [str(x.to_secs()) for x in timestamps2][::step]
    axs[1].set_xticks(ticks=t2[::step])
    axs[1].set_xticklabels(labels=xticks, fontsize=24)
    yticks = np.arange(min(v2), max(v2), 2)
    axs[1].set_yticks(ticks=yticks)
    axs[1].set_yticklabels(labels=['{:.1f}'.format(l) for l in yticks], fontsize=24)

    step = 30
    xticks = [str(x.to_secs()) for x in timestamps3][::step]
    axs[2].set_xticks(ticks=t3[::step])
    axs[2].set_xticklabels(labels=xticks, fontsize=24)
    yticks = np.arange(min(v3), max(v3), 2)
    axs[2].set_yticks(ticks=yticks)
    axs[2].set_yticklabels(labels=['{:.1f}'.format(l) for l in yticks], fontsize=24)

    axs[0].set_xlim(t1[0], t1[-1])
    axs[0].set_xlabel('t [hh:mm]', fontsize=24)
    axs[0].set_ylabel('F [%]', fontsize=24)

    axs[1].set_xlim(t2[0], t2[-1])
    axs[1].set_xlabel('t [hh:mm]', fontsize=24)
    axs[1].set_ylabel('Human Position X [m]', fontsize=24)

    axs[2].set_xlim(t3[0], t3[-1])
    axs[2].set_xlabel('t [hh:mm]', fontsize=24)
    axs[2].set_ylabel('Human Position Y [m]', fontsize=24)

    # axs[0].legend(fontsize=20)
    # axs[1].legend(fontsize=20)

    fig.savefig((SAVE_PATH + '{}.pdf').format(os.environ['LSHA_RES_PATH'], title))

    del fig, axs



from typing import List

DAYS_PER_MONTH = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]


class Timestamp:
    def __init__(self, y: int, m: int, d: int, h: int, min: int, sec: float):
        self.year = y
        self.month = m
        self.day = d
        self.hour = h
        self.min = min
        self.sec = sec

    @staticmethod
    def from_secs(secs: int):
        new_ts = Timestamp(int(secs / (365 * 24 * 3600)), 0, 0, 0, 0, 0)
        if secs > 365 * 24 * 3600:
            secs -= new_ts.year * 365 * 24 * 3600
        sum_d = 0
        for i, month_d in enumerate(DAYS_PER_MONTH):
            sum_d += month_d
            if secs <= sum_d * 24 * 3600:
                new_ts.month = i + 1
                secs -= (sum_d - month_d) * 24 * 3600
                break
        new_ts.day = int(secs / (24 * 3600)) + 1
        secs -= (new_ts.day - 1) * 24 * 3600
        new_ts.hour = int(secs / 3600)
        secs -= new_ts.hour * 3600
        new_ts.min = int(secs / 60)
        new_ts.sec = secs - new_ts.min * 60
        return new_ts

    def to_secs(self):
        months = sum(DAYS_PER_MONTH[:self.month - 1]) if self.month > 0 else 0
        days = self.year * 365 + months + self.day - 1 if self.day > 0 else 0
        minutes = self.hour * 60 + self.min
        seconds = minutes * 60 + self.sec
        return days * 24 * 3600 + seconds

    def __str__(self):
        return '{}/{}/{} {}:{}:{}'.format(self.day, self.month, self.year, self.hour, self.min, self.sec)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return self.to_secs() == other.to_secs()

    def __ge__(self, other):
        return self.to_secs() >= other.to_secs()

    def __lt__(self, other):
        return self.to_secs() < other.to_secs()

    def __sub__(self, other):
        return self.to_secs() - other.to_secs()


class SignalPoint:
    def __init__(self, t: Timestamp, val: float):
        self.timestamp = t
        self.value = val

    def __str__(self):
        return '{}: {}'.format(self.timestamp, self.value)

    def __eq__(self, other):
        return self.timestamp == other.timestamp and self.value == other.value


class SampledSignal:
    def __init__(self, pts: List[SignalPoint], label=None):
        self.label = label
        self.points = pts


class Event:
    def __init__(self, guard, chan, symb):
        self.guard = guard
        self.chan = chan
        self.symbol = symb
        self.label = guard + ',' + chan if len(guard) > 0 else chan

    def __str__(self):
        return self.label

    def __eq__(self, other):
        return self.guard == other.guard and self.chan == other.chan

    def __hash__(self):
        return hash(self.label)


class ChangePoint:
    def __init__(self, t: Timestamp):
        self.t = t
        self.event = None

    def set_event(self, e: Event):
        self.event = e

    def __str__(self):
        return '{} -> {}'.format(self.t, self.event)

    def __eq__(self, other):
        return self.t == other.t and self.event == other.event



from typing import List, Dict

from sha_learning.domain.lshafeatures import Trace, State, EMPTY_STRING
from sha_learning.domain.shafeatures import StochasticHybridAutomaton, Location, Edge
from sha_learning.learning_setup.logger import Logger

LOGGER = Logger('Obs.Table Handler')


class Row:
    def __init__(self, state: List[State]):
        self.state = state

    def is_populated(self):
        return any([s.observed() for s in self.state])

    def __str__(self):
        return '\t|\t'.join([str(s) for s in self.state])

    def __eq__(self, other):
        return all([s == other.state[i] for i, s in enumerate(self.state)])

    def __hash__(self):
        return hash(str(self))


class ObsTable:
    def __init__(self, s: List[Trace], e: List[Trace], low_s: List[Trace]):
        self.__S: List[Trace] = s
        self.__low_S: List[Trace] = low_s
        self.__E: List[Trace] = e
        self.__upp_obs: List[Row] = [Row([State([(None, None)])] * len(e))] * len(s)
        self.__low_obs: List[Row] = [Row([State([(None, None)])] * len(e))] * len(low_s)

    def get_S(self):
        return self.__S

    def add_S(self, word: Trace):
        self.__S.append(word)

    def get_E(self):
        return self.__E

    def add_E(self, word: Trace):
        self.__E.append(word)

    def get_low_S(self):
        return self.__low_S

    def add_low_S(self, word: Trace):
        self.__low_S.append(word)

    def del_low_S(self, index: int):
        self.get_low_S().pop(index)

    def get_upper_observations(self):
        return self.__upp_obs

    def set_upper_observations(self, obs_table: List[Row]):
        self.__upp_obs = obs_table

    def get_lower_observations(self):
        return self.__low_obs

    def set_lower_observations(self, obs_table: List[Row]):
        self.__low_obs = obs_table

    def __str__(self, filter_empty=False):
        result = ''

        rows = self.get_upper_observations() + self.get_lower_observations()
        populated_rows = [i for i, row in enumerate(rows) if row.is_populated()]

        max_tabs = max(
            [len(str(word)) for i, word in enumerate(self.get_S() + self.get_low_S()) if i in populated_rows])
        HEADER = ' ' * max_tabs + '|'

        len_row_cells = [[len(s.label) for s in r.state] for r in rows]
        col_width = [max([l for r in len_row_cells for j_2, l in enumerate(r) if j_2 == j]) for j, e in
                     enumerate(self.get_E())]

        # column (E set) labels
        HEADER += '|'.join([str(e) + ' ' * (col_width[j] - len(str(e))) for j, e in enumerate(self.get_E())])
        result += HEADER + '\n'

        SEPARATOR = '-' * max_tabs + '+' + '+'.join(['-' * c for c in col_width])
        result += SEPARATOR + '\n'

        # print short words row labels
        for i, s_word in enumerate(self.get_S() + self.get_low_S()):
            if i == len(self.get_S()):
                result += SEPARATOR + '\n'
            row = rows[i]
            if filter_empty and not row.is_populated():
                pass
            else:
                ROW = str(s_word)
                ROW += ' ' * (max_tabs - len(str(s_word))) + '|'
                ROW += '|'.join([s.label + ' ' * (col_width[j] - len(s.label)) for j, s in enumerate(row.state)])
                result += ROW + '\n'

        result += SEPARATOR + '\n'
        return result

    def print(self, filter_empty=False):
        LOGGER.info(self.__str__(filter_empty))

    def get_loc_from_word(self, word: Trace, locations: List[Location], seq_to_loc: Dict[Trace, str], teacher):
        loc = None

        if word in seq_to_loc.keys():
            loc = [l for l in locations if l.name == seq_to_loc[word]][0]
        else:
            if word in self.get_S():
                curr_row = self.get_upper_observations()[self.get_S().index(word)]
            elif word in self.get_low_S():
                curr_row = self.get_lower_observations()[self.get_low_S().index(word)]
            elif Trace(word[:-1]) in self.get_S():
                row = self.get_S().index(Trace(word[:-1]))
                column = self.get_E().index(Trace([word[-1]]))
                needed_state = self.get_upper_observations()[row].state[column]
                curr_row = Row([needed_state] + [State([(None, None)])] * (len(self.get_E()) - 1))
            elif Trace(word[:-1]) in self.get_low_S():
                row = self.get_low_S().index(Trace(word[:-1]))
                column = self.get_E().index(Trace([word[-1]]))
                needed_state = self.get_lower_observations()[row].state[column]
                curr_row = Row([needed_state] + [State([(None, None)])] * (len(self.get_E()) - 1))

            if not curr_row.is_populated():
                return loc

            for i, row in enumerate(self.get_upper_observations()):
                if self.get_S()[i] in seq_to_loc.keys() and teacher.eqr_query(curr_row, row, strict=True):
                    loc = [l for l in locations if l.name == seq_to_loc[self.get_S()[i]]][0]
        return loc

    def add_init_edges(self, locations: List[Location], edges: List[Edge], seq_to_loc: Dict[Trace, str], teacher):
        init_loc = Location('__init__', None)
        locations.append(init_loc)

        one_word_upper = [word for word in self.get_S() if len(word) == 1]
        one_word_lower = [word for word in self.get_low_S() if len(word) == 1]

        for word in one_word_upper + one_word_lower:
            dest_loc = self.get_loc_from_word(word, locations, seq_to_loc, teacher)
            if dest_loc is not None:
                edges.append(Edge(init_loc, dest_loc, sync=str(word)))

        return locations, edges

    def to_sha(self, teacher):
        locations: List[Location] = []
        upp_obs: List[Row] = self.get_upper_observations()
        low_obs: List[Row] = self.get_lower_observations()
        # each unique sequence in the upper observations
        # constitutes an automaton location
        unique_sequences: List[Trace] = []
        unique_sequences_dict: Dict[Trace, str] = {}
        for i, row in enumerate(upp_obs):
            row_already_present = False
            for seq in unique_sequences:
                row_2 = upp_obs[self.get_S().index(seq)]
                if teacher.eqr_query(row, row_2, strict=True):
                    row_already_present = True
                    break
            if not row_already_present:
                unique_sequences.append(self.get_S()[i])
        # Create a new location for each unique sequence
        for index, seq in enumerate(unique_sequences):
            seq_index = self.get_S().index(seq)
            row = upp_obs[seq_index]
            new_name = StochasticHybridAutomaton.LOCATION_FORMATTER.format(len(locations))
            new_flow = row.state[0].vars[0][0].label + ', ' + row.state[0].vars[0][1].label
            locations.append(Location(new_name, new_flow))
            unique_sequences_dict[seq] = new_name

        # start building edges list for upper part of the table
        edges: List[Edge] = []
        for s_i, s_word in enumerate(self.get_S()):
            for t_i, t_word in enumerate(self.get_E()):
                if upp_obs[s_i].state[t_i].observed():
                    word: Trace = s_word + t_word
                    entry_word = Trace(word[:len(word) - 1])
                    if len(entry_word) == 0:
                        continue

                    start_loc = self.get_loc_from_word(entry_word, locations, unique_sequences_dict, teacher)
                    dest_loc = self.get_loc_from_word(word, locations, unique_sequences_dict, teacher)

                    labels = str(Trace(word[-1:]))
                    new_edge = Edge(start_loc, dest_loc, sync=labels)
                    if start_loc is not None and dest_loc is not None and new_edge not in edges:
                        edges.append(new_edge)

        # start building edges list for lower part of the table
        for s_i, s_word in enumerate(self.get_low_S()):
            for t_i, t_word in enumerate(self.get_E()):
                if low_obs[s_i].state[t_i].observed():
                    word: Trace = s_word + t_word
                    entry_word = Trace(word[:len(word) - 1])
                    if len(entry_word) == 0:
                        continue

                    start_loc = self.get_loc_from_word(entry_word, locations, unique_sequences_dict, teacher)
                    dest_loc = self.get_loc_from_word(word, locations, unique_sequences_dict, teacher)

                    if word != '':
                        labels = str(word.sub_prefix(entry_word))
                    else:
                        labels = EMPTY_STRING
                    new_edge = Edge(start_loc, dest_loc, sync=labels)
                    if start_loc is not None and dest_loc is not None and new_edge not in edges:
                        edges.append(new_edge)

        locations, edges = self.add_init_edges(locations, edges, unique_sequences_dict, teacher)

        return StochasticHybridAutomaton(locations, edges), unique_sequences_dict



import configparser
import os
from typing import Dict, List

import matplotlib.pyplot as plt
import numpy as np

from sha_learning.domain.lshafeatures import TimedTrace

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

SAVE_PATH = config['SUL CONFIGURATION']['SHA_SAVE_PATH']

SPEED_RANGE = int(config['ENERGY CS']['SPEED_RANGE'])
MIN_SPEED = int(config['ENERGY CS']['MIN_SPEED'])
MAX_SPEED = int(config['ENERGY CS']['MAX_SPEED'])
CS = config['SUL CONFIGURATION']['CASE_STUDY']


def double_plot(timestamps1, v1, timestamps2, v2, t: TimedTrace, title, filtered=False, timestamps3=None, v3=None):
    subplots = 2 if timestamps3 is None else 3
    fig, axs = plt.subplots(subplots, figsize=(60, 30), gridspec_kw={'height_ratios': [3, 3, 3]})

    SIG_WIDTH = 2.0

    t1 = [x.to_secs() for x in timestamps1]
    # t1 = [i for i, x in enumerate(timestamps1)]
    axs[0].plot(t1, v1, 'k-', label='power', linewidth=SIG_WIDTH)
    axs[0].plot(t1, [0] * len(v1), 'k--', linewidth=.5)

    t2 = [x.to_secs() for x in timestamps2]
    # t2 = [i for i, x in enumerate(timestamps2)]
    axs[1].plot(t2, v2, 'k-', label='speed', linewidth=SIG_WIDTH)
    axs[1].plot(t2, [0] * len(v2), 'k-', linewidth=.5)

    if timestamps3 is not None:
        t3 = [x.to_secs() for x in timestamps3]
        # t3 = [i for i, x in enumerate(timestamps3)]
        axs[2].plot(t3, v3, 'k-', label='pressure', linewidth=SIG_WIDTH)
        axs[2].plot(t3, [0] * len(v3), 'k-', linewidth=.5)

    LABEL_FONT = 32
    TICK_FONT = 30
    EVENT_FONT = 22
    EVENT_WIDTH = 2.0
    TITLE_FONT = 38
    MARKER_SIZE = 30

    HEIGHT1 = max(v1) + 1
    HEIGHT2 = max(v2) + 100
    HEIGHT3 = max(v3) + 100

    colors = ['orange', 'b', 'green', 'red']
    labels = ['spindle start', 'spindle stop', 'pressure up', 'pressure down']

    marker = 'x'
    height1 = HEIGHT1
    height2 = HEIGHT2
    height3 = HEIGHT3

    i = 0
    labels = [e.symbol for e in t.e]
    events = [ts.to_secs() for ts in t.t]
    # events = [[i for i in t1 if timestamps1[i].to_secs() == e_t.to_secs()][0] for e_t in t.t]
    axs[0].vlines(events, [0] * len(events), [height1] * len(events), color=colors[i], linewidth=EVENT_WIDTH)
    for i, e in enumerate(events):
        if labels[i] == 'l':
            color = colors[2]
            marker = '^'
        elif labels[i] == 'u':
            color = colors[3]
            marker = 'v'
        elif labels[i] == 'i_0':
            color = colors[1]
            marker = 'v'
        else:
            color = colors[0]
            marker = '^'
        axs[0].plot(e, height1, marker, color=color, label=labels[i], markersize=MARKER_SIZE)
        axs[0].vlines(e, 0, height1, color='k', linewidth=EVENT_WIDTH)
        # axs[0].text(e, height1, labels[i], fontsize=EVENT_FONT)

    # axs[0].plot(events[1:-1], [0] * (len(events) - 2), color='k', linewidth=50, zorder=1)
    # ops = ['26', '14', '16', '4', '26', '2']
    # i_op = 0
    i = 0
    # axs[2].plot(events, [height2] * len(events), marker, color=colors[i], label=labels[i])
    # axs[2].vlines(events, [0] * len(events), [height2] * len(events), color=colors[i], linewidth=EVENT_WIDTH)
    for i, e in enumerate(events):
        if labels[i] in ['l', 'u']:
            continue

        if labels[i] == 'i_0':
            color = colors[1]
            marker = 'v'
        else:
            color = colors[0]
            marker = '^'
        axs[1].plot(e, height2, marker, color=color, label=labels[i], markersize=MARKER_SIZE)
        axs[1].vlines(e, 0, height2, color='k', linewidth=EVENT_WIDTH)
        # axs[0].vlines(e, -0.1, +0.1, color='white', linewidth=8, zorder=4)
        # if labels[i] not in ['l', 'u', 'i_0']:
        # print('{}: {}'.format(ops[i_op], e + (events[i + 1] - e)))
        # axs[0].text(e + (events[i + 1] - e) / 2, -0.05, ops[i_op], color='white', zorder=4, fontsize=40)
        # i_op += 1
        # axs[2].text(e, height2, labels[i], fontsize=EVENT_FONT)

    i = 0
    # axs[3].plot(events, [height3] * len(events), marker, color=colors[i], label=labels[i])
    # axs[3].vlines(events, [0] * len(events), [height3] * len(events), color=colors[i], linewidth=EVENT_WIDTH)
    for i, e in enumerate(events):
        if labels[i].startswith('i') or labels[i].startswith('m'):
            continue
        if labels[i] == 'l':
            color = colors[2]
            marker = '^'
        else:
            color = colors[3]
            marker = 'v'
        axs[2].plot(e, height3, marker, color=color, label=labels[i], markersize=MARKER_SIZE)
        axs[2].vlines(e, 0, height3, color='k', linewidth=EVENT_WIDTH)
        # axs[3].text(e, height3, labels[i], fontsize=EVENT_FONT)

    PAD = 0.1

    step = 300
    xticks = [str(x.hour) + ':' + str(x.min).zfill(2) for x in timestamps1][::step]
    # xticks = [str(x) for x in t1][::step] + [str(t1[-1])]
    # axs[1].set_xticks(ticks=[int(o) for o in xticks])
    axs[0].set_xticks(ticks=[x for x in t1[::step]])
    axs[0].set_xticklabels(labels=xticks, fontsize=TICK_FONT)
    xmin, xmax = axs[0].get_xlim()
    axs[0].set_xlim(xmin - PAD, xmax)
    yticks = np.arange(0, max(v1) + 1, 1)
    axs[0].set_yticks(ticks=yticks)
    axs[0].set_yticklabels(labels=yticks, fontsize=TICK_FONT)
    ymin, ymax = axs[0].get_ylim()
    axs[0].set_ylim(0, ymax)

    # step = 60
    xticks = [str(x.hour) + ':' + str(x.min).zfill(2) for x in timestamps2][::step]
    # axs[2].set_xticks(ticks=[int(o) for o in xticks])
    axs[1].set_xticks(ticks=[x for x in t1[::step]])
    axs[1].set_xticklabels(labels=xticks, fontsize=TICK_FONT)
    xmin, xmax = axs[1].get_xlim()
    axs[1].set_xlim(xmin - PAD, xmax)
    yticks = np.arange(0, max(v2) + 400, 400)
    axs[1].set_yticks(ticks=yticks)
    axs[1].set_yticklabels(labels=yticks, fontsize=TICK_FONT)
    ymin, ymax = axs[1].get_ylim()
    axs[1].set_ylim(0, ymax)

    # step = 120
    xticks = [str(x.hour) + ':' + str(x.min).zfill(2) for x in timestamps3][::step]
    # axs[3].set_xticks(ticks=[int(o) for o in xticks])
    axs[2].set_xticks(ticks=[x for x in t1[::step]])
    axs[2].set_xticklabels(labels=xticks, fontsize=TICK_FONT)
    xmin, xmax = axs[2].get_xlim()
    axs[2].set_xlim(xmin - PAD, xmax)
    yticks = np.arange(0, max(v3) + 1, 800)
    axs[2].set_yticks(ticks=yticks)
    # axs[2].set_yticklabels(labels=['unlocked', 'locked'], fontsize=TICK_FONT)
    ymin, ymax = axs[2].get_ylim()
    axs[2].set_ylim(0, ymax)

    # axs[0].set_xticks(ticks=[int(o) for o in xticks])
    # axs[0].set_xticklabels(labels=xticks, fontsize=TICK_FONT)
    # axs[0].set_yticks(ticks=[0])
    # axs[0].set_yticklabels(labels=['OP'], fontsize=TICK_FONT)
    # axs[0].set_xlabel('t [s]', fontsize=LABEL_FONT)
    # axs[0].set_ylabel('', fontsize=LABEL_FONT)
    # axs[0].set_title('Operation ID', fontsize=TITLE_FONT)

    # axs[1].set_xlim(t1[0], t1[-1])
    axs[0].set_xlabel('t [hh:mm]', fontsize=LABEL_FONT)
    axs[0].set_ylabel('[kW]', fontsize=LABEL_FONT)
    axs[0].set_title('Total Power (spindle power + axis power)', fontsize=TITLE_FONT)

    # axs[2].set_xlim(t1[0], t1[-1])
    axs[1].set_xlabel('t [hh:mm]', fontsize=LABEL_FONT)
    axs[1].set_ylabel('[rpm]', fontsize=LABEL_FONT)
    axs[1].set_title('Spindle Speed', fontsize=TITLE_FONT)

    # axs[3].set_xlim(t1[0], t1[-1])
    axs[2].set_xlabel('t [hh:mm]', fontsize=LABEL_FONT)
    axs[2].set_ylabel('', fontsize=LABEL_FONT)
    axs[2].set_title('Clamping Pressure', fontsize=TITLE_FONT)

    # axs[1].legend(fontsize=20)
    # axs[2].legend(fontsize=20)

    plt.tight_layout(pad=10.0)
    fig.savefig(SAVE_PATH + '{}.pdf'.format(title))

    del fig, axs


def distr_hist(values: Dict[int, List[float]], name: str):
    values = [(v, values[v]) for v in values]
    # values = sorted(values, key=lambda tup: sum(tup[1]) / len(tup[1]))

    fig, axs = plt.subplots(1, len(values), tight_layout=True, figsize=(5 * len(values), 5))

    for i, ax in enumerate(axs):
        ax.set_title('D_{}'.format(i))
        if CS == 'THERMO':
            ax.hist(values[i][1], bins=25)
        else:
            ax.hist(values[i][1], bins=25, density=True, histtype='step')
            with open(SAVE_PATH + '{}.txt'.format('histogram_values'), 'a') as f:
                f.write('D_{}:\n'.format(i))
                lines = [str(x) + '\n' for x in values[i][1]]
                print(lines)
                f.writelines(lines)
    fig.savefig(SAVE_PATH + '{}_{}.pdf'.format(name, 'histograms'))
    del fig
